//#define NV_OLD
#define OUTPUT_MULTIPLE_FILES
#define DONT_DUMP_CHANGED_REGS
#define DONT_DUMP_ABOVE_40000
//#define SIMPLE_FIFO_OUTPUT


#include "re.h"
#include "nv_hw.h"
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <string.h>
#include <SDL/SDL_syswm.h>
#include <errno.h>
#include "NVCtrl.h"
#include "NVCtrlLib.h"
#include "objects.h"
#include "regl.h"
#include "detection.h"

#define MAPPING_FIFO	1
#define MAPPING_REGS	2
#define MAPPING_NODUMP	4
#define MAPPING_AGP     8


static mapping mappings[100];

static int ignore_regs[]={0x25,0x3b,0x15c,0x15d,0x15e,0x16d};

/* run_xvmc_tests wants the maps to be re-read, so it fiddles with init_ok */
int init_ok = 0;

static long agp_base = 0;
static long agp_size = 0;

unsigned int* fb = NULL;
unsigned int  fb_size;

unsigned int* all_regs = NULL;
unsigned int card_family = 0;
int          card_type   = -1;


/* Initialize SDL for calling SDL_SetVideoMode with OpenGL.
 * Called from main() and detect_devices().
 */
void setup_SDL(int opengl)
{
	SDL_Init(SDL_INIT_VIDEO);
	if (opengl) {
		//SDL_GL_SetAttribute(SDL_GL_STEREO,1);
		SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER,0);
		SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE,16);
		SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE, 8);
	}
}

unsigned int *
gpu_ofs_to_agp(unsigned long ofs)
{
	unsigned long ret = 0;
	int i;

	for (i=0;mappings[i].start != NULL;i++) {
		if (mappings[i].type & MAPPING_AGP) {
			unsigned long addr = ofs + agp_base;
			if( (addr >= mappings[i].phys) &&
			    (addr < (mappings[i].phys + mappings[i].size)) )
			{
				
				ret  = (ofs - (mappings[i].phys - agp_base));
				ret += (unsigned long)mappings[i].start;
				break;
			}
		}
	}

	return (unsigned int *)ret;
}

static int must_ignore(int r)
{
	int i;
	for(i=0;i<sizeof(ignore_regs)/sizeof(int);i++)
		if (ignore_regs[i]==r)
			return 1;
	return 0;
}

static int regcmp(unsigned int* m1,unsigned int* m2,int size)
{
	int i;
	for(i=0;i<size/4;i++)
	{
		if (must_ignore(i))
			continue;
		if (m1[i]!=m2[i])
			return 1;
	}
	return 0;
}

char *bit_pattern(unsigned int val)
{
    static char patbuf[40];
    int count = 32;
    char *p = &patbuf[39];

    *p = '\0';
    p--;

    while (count--) {

        if (val & 1)
            *p = '1';
        else
            *p = '0';

        p--;
        val >>= 1;

        if (!(count % 4)) {
            *p = ' ';
            p--;
        }
    }

    return patbuf;
}

int get_vidram()
{
	if (card_family==NV04)
	{
			int config=all_regs[0x00100000/4]&0x3;
			int amount[4]={32,4,8,16};
			return amount[config];
	} else {
			// won't work on nforces. does anyone have an nforce ?
			return (all_regs[0x0010020c/4]&0xfff00000)>>20;
	}		
}

static unsigned int get_card_family(uint32_t card_id)
{
	switch(card_id & 0x0FF0)
	{
		case 0x0020: /* TNT1/TNT2 */
		case 0x00A0: /* Aladdin TNT2 */
			return NV04;
		case 0x0100: /* GeForce 256 */
			return NV10;
		case 0x0110: /* GeForce2 MX */
		case 0x01A0: /* nForce */
		case 0x0150: /* GeForce2 */
		case 0x0170: /* GeForce4 MX */
		case 0x0180: /* GeForce4 MX (8x AGP) */
		case 0x01F0: /* nForce2 */
			return NV15;
		case 0x0200: /* GeForce3 */
		case 0x0250: /* GeForce4 Ti */
		case 0x0280: /* GeForce4 Ti (8x AGP) */
			return NV20;
		case 0x0300: /* GeForceFX 5800 */
		case 0x0310: /* GeForceFX 5600 */
		case 0x0320: /* GeForceFX 5200 */
		case 0x0330: /* GeForceFX 5900 */
		case 0x0340: /* GeForceFX 5700 */
			return NV30;
		case 0x0040: /* GeForce 6800 */
		case 0x00C0: /* GeForce 6800 */
		case 0x0120: /* GeForce 6800 */
		case 0x0210: /* GeForce 6800 */
		case 0x0140: /* GeForce 6600 */
		case 0x0160: /* GeForce 6200 */
		case 0x0220: /* GeForce 6200 */
		case 0x0240: /* GeForce 6100 */
			return NV40;
		case 0x01D0: /* GeForce 7200, 7300, 7400 */
		case 0x0090: /* GeForce 7800 */
		case 0x0290: /* GeForce 7900 */
		case 0x02E0: /* GeForce 7600 */
		case 0x0390: /* GeForce 7600 */
			return G70;
		case 0x00F0: /* misc PCI-E cards */
			switch(card_id&0xFF)
			{
				case 0xFF: /* Geforce 4300 */
					return NV15;
				case 0xFA: /* GeForceFX 5750 */
				case 0xFB: /* GeForceFX 5900 */
				case 0xFC: /* QuadroFX 330/GeForceFX 5300 */
				case 0xFD: /* QuadroFX 330/NVS 280 */
				case 0xFE: /* QuadroFX 1300 */
					return NV30;
				case 0xF0: /* GeForce 6800 */
				case 0xF1: /* GeForce 6600 */
				case 0xF2: /* GeForce 6600 */
				case 0xF3: /* GeForce 6200 */
				case 0xF4: /* GeForce 6600 */
				case 0xF6: /* GeForce 6600 */
				case 0xF7: /* guess */
				case 0xF8: /* QuadroFX 3400/4400 */
				case 0xF9: /* GeForce 6800 */
					return NV40;
				case 0xF5: /* GeForce 7800 */
					return G70;
			}
			break;
		case 0x190:
			return G80;
		default:
			printf("Unknown card\n");
			return G70;
	}
	return -1;
}

static int query_bus_type()
{
	int btype = -1;
	/* call NV_CONTROL to find the bus type */
	SDL_SysWMinfo sdli;
	SDL_VERSION(&sdli.version);
	int nvc_base, nvc_err;
	if (SDL_GetWMInfo(&sdli))
	{
		int ok;
		ok = XNVCTRLQueryExtension(sdli.info.x11.display, &nvc_base,
		                           &nvc_err);
		if (ok)
		{
			if (!XNVCTRLQueryAttribute(sdli.info.x11.display, 0,
			      0, NV_CTRL_BUS_TYPE, &btype))
			{
				printf("\nNV-CONTROL query failed.\n");
				return -1;
			}
		} else
			printf("\nNV-CONTROL extension not found.\n");
	} else
		printf("\nFailed to get SDL_SysWMInfo\n");

	return btype;
}

static int memvalid(const void* addr)
{
	int fd=open("/dev/random",O_WRONLY);
	int r=write(fd,addr,1);
	close(fd);
	return (r==1);
}

void init_card()
{
	if (all_regs != NULL)
		return;

	/* first, figure out the physical adress for the regs
	 * These were detected in find_pci_device().
	 */
	unsigned phys_addr, card_id, fb_phys;

	card_id = (pcidev_->vendor << 16) | pcidev_->device;
	phys_addr = pcidev_->resource[0].start;
	fb_phys = pcidev_->resource[1].start & 0xFFFFFFF0;
	fb_size = pcidev_->resource[1].len;

	// now we can map it
	int fd;
	fd=open(device_node_, O_RDWR);
	if (fd<0) {
		perror("open");
		exit(1);
	}
	all_regs = mmap(NULL, ALL_REGS_SIZE, PROT_READ|PROT_WRITE,
	                MAP_SHARED, fd,
#ifdef NV_OLD
			0
#else
			phys_addr
#endif
			);
	if (all_regs == MAP_FAILED) {
		close(fd);
		perror("mmap");
		exit(1);
	}
	fb = mmap(NULL, fb_size, PROT_READ|PROT_WRITE, MAP_SHARED, fd, fb_phys);
	if (fb == MAP_FAILED) {
		printf("Failed to map video memory!\n");
		fb = NULL;
	}
	close(fd);

	// now figure out the card family
	card_family = get_card_family(card_id);

	printf("Card has id 0x%x ",card_id);
	switch(card_family)
	{
		case NV04:
			printf("(NV04) ");
			break;
		case NV10:
			printf("(NV10) ");
			break;
		case NV15:
			printf("(NV15) ");
			break;
		case NV20:
			printf("(NV20) ");
			break;
		case NV30:
			printf("(NV30) ");
			break;
		case NV40:
			printf("(NV40) ");
			break;
		case G70:
			printf("(G70) ");
			break;
		case G80:
			printf("(G80) ");
			break;
		default:
			printf("(unknown) ");
			break;
	}
#ifdef NV_OLD
	card_type = NV_CTRL_BUS_TYPE_AGP; // sane default :)
#else
	card_type = query_bus_type();
#endif
	switch(card_type)
	{
		case NV_CTRL_BUS_TYPE_AGP:
			printf("bus is AGP");
			break;
		case NV_CTRL_BUS_TYPE_PCI:
			printf("bus is PCI");
			break;
		case NV_CTRL_BUS_TYPE_PCI_EXPRESS:
			printf("bus is PCI-E");
			break;
		case NV_CTRL_BUS_TYPE_INTEGRATED:
			printf("bus is integrated");
			break;
		default:
			printf("unknown bus type");
			card_type=-1;
			break;
	}
	printf(" (%d MB video RAM)",get_vidram());
	printf(" BOOT0 is 0x%08x\n",all_regs[0]);

	// try to find the address/size of the AGP aperture
	if (card_type == NV_CTRL_BUS_TYPE_AGP)
	{
		pci_resource res;
		if (find_agp_aperture(&res) == 0)
		{
			agp_base = res.start & ~8; // wtf is this?
			agp_size = res.len;
			printf("AGP aperture @ 0x%lX, %ldMB\n", agp_base,
			       agp_size/1024/1024);
		} else
		{
			agp_base = agp_size = 0;
			printf("Couldn't determine address of AGP aperture\n");
		}
	}

	shader_init_funcs();
	// figure out which fifos are in use
	/*
	unsigned fifos=all_regs[NV_FIFO_ENABLE/4];
	printf("Fifos state : %x\n",fifos);
	{
		int fd;
		printf("mapping\n");
		fd=open(device_node_, O_RDWR);
		if (mmap(NULL, FIFO_SIZE1, PROT_READ, MAP_SHARED, fd,
		         0xe0011000-2*0x8000) == MAP_FAILED) {
			perror("mmap");
			exit(1);
		}
		close(fd);
	}
	*/
}


static void find_maps()
{
	char s[1024];
	int mapnum=0;

	/* before we read the maps, we create one more mapping
	 * with all the regs */
	init_card();

	FILE *f = fopen("/proc/self/maps", "r");
	while(fgets(s,1024,f))
	{
		if(parse_mapping(s, &mappings[mapnum]))
		{
			ERR_PRINT("Failed to parse line: %s\n", s);
			continue;
		}

		if(!is_wanted_mapping(&mappings[mapnum]))
		{
			/*ERR_PRINT("Discarding line: %s\n", s);*/
			destroy_mapping(&mappings[mapnum]);
			continue;
		}
		
		mappings[mapnum].before = malloc(mappings[mapnum].size);
		mappings[mapnum].after = malloc(mappings[mapnum].size);

		printf("map %d\tat %p size 0x%lx\t(physical 0x%lx)\t",
		       mapnum, mappings[mapnum].start, mappings[mapnum].size,
		       mappings[mapnum].phys);
				
		if ( (mappings[mapnum].size == FIFO_SIZE1) ||
		     (mappings[mapnum].size == FIFO_SIZE2) ||
		     (mappings[mapnum].size == FIFO_SIZE3) ||
		     (mappings[mapnum].size == FIFO_SIZE4) )
		{
			// the fifo
			mappings[mapnum].type |= MAPPING_FIFO;
			printf("=> fifo\n");
		}
		else if ( (mappings[mapnum].size >= 640*480*2) &&
		          (mappings[mapnum].size != REGS_SIZE) &&
		          (mappings[mapnum].size != WTF_SIZE) )
		{
			// the screen or something else we don't want to dump
			mappings[mapnum].type |= MAPPING_NODUMP;
			printf("=> no dump\n");
		} else if ( (mappings[mapnum].phys >= agp_base) &&
		            (mappings[mapnum].phys < (agp_base + agp_size)) )
		{
			mappings[mapnum].type |= MAPPING_AGP;
			printf("=> AGP\n");
		} else if ( (card_type == NV_CTRL_BUS_TYPE_PCI_EXPRESS) &&
		            ((mappings[mapnum].size == REGS_SIZE2)
			     ||(mappings[mapnum].size == REGS_SIZE3))
			     )
		{
			// crashes when accessed on pci-e
			mappings[mapnum].type |= MAPPING_NODUMP;
			printf("=> registers (no read, PCI-E)\n");
		} else
		{	
			// probably some regs
			mappings[mapnum].type |= MAPPING_REGS;
			printf("=> registers\n");
		}
		
		mapnum++;

		// insane quadro hack
		if (mappings[mapnum-1].size == FIFO_SIZE_QUADRO2)
		{
#define QUADRO_FIFO_OFFSET 0xbf400
			mappings[mapnum].start = mappings[mapnum-1].start + QUADRO_FIFO_OFFSET;
			mappings[mapnum].phys = mappings[mapnum-1].phys + QUADRO_FIFO_OFFSET;
			mappings[mapnum].size = FIFO_SIZE2;
			mappings[mapnum].before = (unsigned int*)malloc(mappings[mapnum].size);
			mappings[mapnum].after = (unsigned int*)malloc(mappings[mapnum].size);
			mappings[mapnum].type = MAPPING_FIFO;
			printf("map %d\tfrom 0x%lx to 0x%lx size 0x%lx"
			       " (physical 0x%lx)\t", mapnum,
			       (unsigned long)mappings[mapnum].start,
			       (unsigned long)(mappings[mapnum].start
			                     + mappings[mapnum].size),
			        mappings[mapnum].size, mappings[mapnum].phys);
			printf("=> quadro fifo\n");
			mapnum++;
		}
	}
	mappings[mapnum].start=NULL;

	clean_dmarecs();

	fclose(f);
}



void print_fifo (int n, int before, int after)
{
#ifdef SIMPLE_FIFO_OUTPUT
    printf("0x%.8x   ", after);
#else
    printf("%x   0x%.8x   0x%.8x   ", n, before, after);
#endif
}

#define ALLREG_START 0x2000
#define ALLREG_SIZE 0x1000
static int32_t allregmapping_before[ALLREG_SIZE/4];
static int32_t allregmapping_after[ALLREG_SIZE/4];

void dumpreg_before()
{
	if (!init_ok)
	{
		find_maps();
		init_ok=1;
	}

	unsigned char *allreg_src = (unsigned char *)all_regs;
	memcpy(allregmapping_before,&allreg_src[ALLREG_START],ALLREG_SIZE);
}

void dumpreg_after()
{
	unsigned char *allreg_src = (unsigned char *)all_regs;
	memcpy(allregmapping_after,&allreg_src[ALLREG_START],ALLREG_SIZE);

	int j;
	for(j=0;j<ALLREG_SIZE/4;j++)
	{
		if (allregmapping_before[j]!=allregmapping_after[j])
		{
			printf("Changed reg 0x%.8x from 0x%.8x to 0x%.8x\n",
			       ALLREG_START+(j<<2), allregmapping_before[j],
			       allregmapping_after[j]);
		}
	}
}

void dump_before()
{
	int i;
	if (!init_ok)
	{
		find_maps();
		init_ok=1;
	}
	i=0;
	// dump each mapping to its buffer
	while(mappings[i].start)
	{
		if (!memvalid(mappings[i].start))
			mappings[i].type|=MAPPING_NODUMP;
		if (!(mappings[i].type&MAPPING_NODUMP))
			memcpy(mappings[i].before,mappings[i].start,mappings[i].size);
		i++;
	}
}


typedef struct {
	unsigned int name;
	unsigned int type;
} fifo_descriptor;

static fifo_descriptor fifo_channels[8];

static void set_subchannel(int channel, unsigned int name) {
	if (!is_name_dumped(name)) {
		printf("mapped subchannel %d to %08x\n", channel, name);
	}
	fifo_channels[channel].name = name;
	fifo_channels[channel].type = find_object_type(name);
	set_name_dumped(name);
}

const char *dma_opcode_name[4] = {
	"METHOD",
	"JUMP",
	"NONINC_METHOD",
	"CALL",
};

static void dump_changed_regs(mapping *map)
{
	int j;
	if( (map->type & MAPPING_REGS) ||
	    (map->type & MAPPING_AGP) )
	{
		// compare the regs, find the changes
		for(j=0; j<map->size/4; j++)
		{
			if (must_ignore(j))
				continue;
			if (map->before[j] == map->after[j])
				continue;
			printf("Changed reg 0x%.8x from 0x%.8x to 0x%.8x\n",
			       j, map->before[j], map->after[j]);
		}
	}
}

static void dump_fifo(mapping *map, int dump_startup)
{
	int j;
	unsigned int size, channel, offset, opcode;
	int changed_from = 0x7FFFFFFF, changed_to = -1;
	int upbound;

	if( !(map->type & MAPPING_FIFO) )
		return;

	if (dump_startup)
		changed_from = 0;

#ifdef NV_OLD
	if (map->size == FIFO_SIZE4)
		upbound = 328000;
		/* yet another hack to avoid huge dumps with
		 * 97xx drivers... */
	else
		upbound = map->size;
#else
	upbound = 0x100000;
#endif

	// compare the fifos, find the changes
	for(j=0; j<upbound/4-1; j++)
	{
		if (map->after[j] != map->before[j])
		{
			if (j<changed_from)
				changed_from=j;
			if (j>changed_to)
				changed_to=j;
		}
	}
	changed_to++;
	if (changed_from>changed_to)
	{
		printf("No fifo changes\n");
		return;
	}

	printf("Changed from %d to %d   (%d words   %d bytes)\n",
	       changed_from, changed_to, (changed_to-changed_from),
	       (changed_to-changed_from)*4);

	j = 0;
	while (j < changed_from)
	{
		size = (map->after[j] & 0x1ffc0000) >> 18;
		channel = (map->after[j] & 0x0000e000) >> 13;
		offset = map->after[j] & 0x1FFC;
		opcode = map->after[j] >> 29;
		if (offset == 0 && size > 0)
			set_subchannel(channel, map->after[j+1]);

		j += size + 1;
	}

	int nops = 0;
	size = 0;
	for(j=changed_from; j<changed_to || size; j++)
	{
		if (size)
		{
			if (offset == 0) {
				set_subchannel(channel, map->after[j]);
			}

			char *str = format_load(fifo_channels[channel].type,
			                        offset, map->after[j], card_family);
			print_fifo(j, map->before[j], map->after[j]);
			if (offset != 0) {
				printf(str);
			}
			printf("\n");
			size--;
			if (opcode != 2 /* NONINC_METHOD */)
				offset += 4;
			continue;
		}
		
		if (map->after[j] == 0x0) {
			nops ++;
			continue;
		}
						
		if (nops != 0) {
			printf("\t\t%d NOPs skipped\n", nops);
			nops = 0;
		}

#ifdef DONT_DUMP_ABOVE_40000
		if (j >= 0x40000) {
			continue;
		}
#endif

		print_fifo(j, map->before[j], map->after[j]);

		size = (map->after[j] & 0x1ffc0000) >> 18;
		channel = (map->after[j] & 0x0000e000) >> 13;
		offset = map->after[j] & 0x1FFC;
		opcode = map->after[j] >> 29;
		printf("          {size: 0x%x   channel: 0x%x   "
		       "obj: %08x opcode: %s offset: %04x}\n", size,
		       channel, fifo_channels[channel].name,
		       (opcode<4) ? dma_opcode_name[opcode] : "UNKNOWN",
		       offset);
	}
}

void dump_after(int dump_startup)
{
	int i;
	if (regl.Flush)
		regl.Flush();

	// dump each mapping to its buffer
	i=0;
	while(mappings[i].start)
	{
		if (!memvalid(mappings[i].start))
			mappings[i].type|=MAPPING_NODUMP;
		if (!(mappings[i].type & MAPPING_NODUMP))
			memcpy(mappings[i].after, mappings[i].start,
			       mappings[i].size);
		i++;
	}
	i=0;
	while(mappings[i].start)
	{
		if( mappings[i].type & MAPPING_NODUMP )
		{
			i++;
			continue;
		}

		if( regcmp(mappings[i].after, mappings[i].before,
		           mappings[i].size) == 0 )
		{
			i++;
			continue;
		}

		printf("==========================\n");
		printf("Mapping %d ",i);
		if (mappings[i].type&MAPPING_REGS)
			printf("(regs)");
		if (mappings[i].type&MAPPING_FIFO)
			printf("(fifo)");
		if (mappings[i].type&MAPPING_NODUMP)
			printf("(screen)");

		printf("\n");

#ifndef DONT_DUMP_CHANGED_REGS
		dump_changed_regs(&mappings[i]);
#endif
		dump_fifo(&mappings[i], dump_startup);

		i++;
		printf("\n");
	}
}

static int _old_fd = -1;
static fpos_t _old_fpos;
void redirect_output(const char* name)
{
#ifdef OUTPUT_MULTIPLE_FILES
	char scratch[1024];

	fflush(stdout);
	fgetpos(stdout, &_old_fpos);
	
	/* preserve the old output target */
	_old_fd = dup(fileno(stdout));
	
	if(pcidev_)
	{
		snprintf(scratch, 1023, "card_%04x-%04x_%s.txt",
		         pcidev_->vendor, pcidev_->device, name);
	} else
	{
		snprintf(scratch, 1023, "card_%s.txt", name);
	}
	
	if(!freopen(scratch, "w", stdout))
	{
		ERR_PRINT("Opening '%s' failed: %s\n",
			scratch, strerror(errno));
	}
#endif
}

/* NOTE: This can be called only once after a redirect_output(),
 * they do not really nest.
 */
void close_output()
{
	fflush(stdout);
#ifdef OUTPUT_MULTIPLE_FILES
	if(-1 == _old_fd)
	{
		ERR_PRINT("no previous stream to return output to, "
		          "continuing with the current output.\n");
		return;
	}

	/* return the previous output file */
	if(dup2(_old_fd, fileno(stdout)) == -1)
		ERR_PRINT("Redirecting to previous file failed: %s\n",
			strerror(errno));
	else
		close(_old_fd);
	_old_fd = -1;
	clearerr(stdout);
	fsetpos(stdout, &_old_fpos);
#endif
}

void let_it_snow()
{
	int i;
	int s;
	i=0;
	s=0;
	while(mappings[i].start)
	{
		if (mappings[i].type&MAPPING_NODUMP)
			s=i;
		i++;
	}
	printf("screen is %d\n",s);
	for(i=0;i<10000000;i++)
	{
		int addr=(rand()%mappings[s].size)&(~0x03);
		*((char*)mappings[s].start+addr)=255;// blue
		*((char*)mappings[s].start+addr+1)=255;// green
		*((char*)mappings[s].start+addr+2)=255;// red
	}
}


